/*
Print Scrn key remap driver

based on : Ctrl2cap Keyboard Filter
Copyright (C) 1996-1999 Mark Russinovich
http://www.sysinternals.com

modified by rpy3uH  (26.06.2011)

E-mail: gruzino.abhaz@gmail.com
*/
#include <ntddk.h>
#include <ntddkbd.h>

#define PRINT_SCREEN   ((USHORT)0x37)
#define ENTER      	   ((USHORT)0x1C)
	
typedef struct _DEVICE_EXTENSION {
    PDEVICE_OBJECT  TopOfStack;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

NTSTATUS
remapkeyAddDevice(
    IN PDRIVER_OBJECT Driver,
    IN PDEVICE_OBJECT PDO)
{
    PDEVICE_EXTENSION   devExt;
    IO_ERROR_LOG_PACKET errorLogEntry;
    PDEVICE_OBJECT      device;
    NTSTATUS            status = STATUS_SUCCESS;

    // Create a filter device and attach it to the device stack.
    status = IoCreateDevice(Driver, sizeof(DEVICE_EXTENSION), 
                            NULL, FILE_DEVICE_KEYBOARD,   
                            0, FALSE, &device);

    if (!NT_SUCCESS(status)) return status;

    RtlZeroMemory(device->DeviceExtension, sizeof(DEVICE_EXTENSION));

    devExt = (PDEVICE_EXTENSION) device->DeviceExtension;
    devExt->TopOfStack = IoAttachDeviceToDeviceStack(device, PDO);

    device->Flags |= (DO_BUFFERED_IO | DO_POWER_PAGABLE);
    device->Flags &= ~DO_DEVICE_INITIALIZING;
    return status;
}

NTSTATUS remapkeyPnP(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp)
{
    PDEVICE_EXTENSION           devExt; 
    PIO_STACK_LOCATION          irpStack;
    NTSTATUS                    status = STATUS_SUCCESS;
    KIRQL                       oldIrql;
    KEVENT                      event;        

    devExt = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    switch (irpStack->MinorFunction) {
    case IRP_MN_REMOVE_DEVICE:
        
        //
        // Detach from the target device after passing the IRP
        // down the devnode stack.
        //
        IoSkipCurrentIrpStackLocation(Irp);
        IoCallDriver(devExt->TopOfStack, Irp);

        IoDetachDevice(devExt->TopOfStack); 
        IoDeleteDevice(DeviceObject);

        status = STATUS_SUCCESS;
        break;

    case IRP_MN_SURPRISE_REMOVAL:

        // Same as a remove device, but don't call IoDetach or IoDeleteDevice.
        IoSkipCurrentIrpStackLocation(Irp);
        status = IoCallDriver(devExt->TopOfStack, Irp);
        break;

    case IRP_MN_START_DEVICE: 
    case IRP_MN_QUERY_REMOVE_DEVICE:
    case IRP_MN_QUERY_STOP_DEVICE:
    case IRP_MN_CANCEL_REMOVE_DEVICE:
    case IRP_MN_CANCEL_STOP_DEVICE:
    case IRP_MN_FILTER_RESOURCE_REQUIREMENTS: 
    case IRP_MN_STOP_DEVICE:
    case IRP_MN_QUERY_DEVICE_RELATIONS:
    case IRP_MN_QUERY_INTERFACE:
    case IRP_MN_QUERY_CAPABILITIES:
    case IRP_MN_QUERY_DEVICE_TEXT:
    case IRP_MN_QUERY_RESOURCES:
    case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:
    case IRP_MN_READ_CONFIG:
    case IRP_MN_WRITE_CONFIG:
    case IRP_MN_EJECT:
    case IRP_MN_SET_LOCK:
    case IRP_MN_QUERY_ID:
    case IRP_MN_QUERY_PNP_DEVICE_STATE:
    default:
        // Pass these through untouched
        IoSkipCurrentIrpStackLocation(Irp);
        status = IoCallDriver(devExt->TopOfStack, Irp);
        break;
    }

    return status;
}

NTSTATUS remapkeyPower(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp)
{
    PDEVICE_EXTENSION   devExt;
    
    devExt = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
    PoStartNextPowerIrp(Irp); 
    IoSkipCurrentIrpStackLocation(Irp);
    
    return PoCallDriver( devExt->TopOfStack, Irp );
}

NTSTATUS remapkeyReadComplete( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp,
    IN PVOID Context)
{
    PIO_STACK_LOCATION        IrpSp;
    PKEYBOARD_INPUT_DATA      KeyData;
    LONG numKeys, i;

    IrpSp = IoGetCurrentIrpStackLocation(Irp);
    if(NT_SUCCESS(Irp->IoStatus.Status)) 
    {
        KeyData = Irp->AssociatedIrp.SystemBuffer;
        numKeys = Irp->IoStatus.Information / sizeof(KEYBOARD_INPUT_DATA);

        for( i = 0; i < numKeys; i++ ) 
        {
       		if (KeyData[i].MakeCode == PRINT_SCREEN)
       		{
       			KeyData[i].MakeCode = ENTER;
       		}
        }
    }
    
    if (Irp->PendingReturned) IoMarkIrpPending( Irp );

    return Irp->IoStatus.Status;
}

NTSTATUS 
remapkeyDispatchRead( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp )
{
    PDEVICE_EXTENSION   devExt;
    PIO_STACK_LOCATION  currentIrpStack;
    PIO_STACK_LOCATION  nextIrpStack;

    devExt = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
    currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    nextIrpStack = IoGetNextIrpStackLocation(Irp);    

    *nextIrpStack = *currentIrpStack;
   
    IoSetCompletionRoutine( Irp, remapkeyReadComplete, 
                            DeviceObject, TRUE, TRUE, TRUE );

    return IoCallDriver( devExt->TopOfStack, Irp );
}

NTSTATUS 
DriverDispatchGeneral(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp)
{
    IoSkipCurrentIrpStackLocation(Irp);
    return IoCallDriver(((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->TopOfStack, Irp);
}

void DriverUnload(IN PDRIVER_OBJECT Driver)
{
    UNREFERENCED_PARAMETER(Driver);
}

NTSTATUS DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath)
{
    ULONG i;
    NTSTATUS status;

    for (i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++) 
    {
        DriverObject->MajorFunction[i] = DriverDispatchGeneral;
    }

    DriverObject->MajorFunction[IRP_MJ_READ] = remapkeyDispatchRead;

    DriverObject->MajorFunction [IRP_MJ_POWER]  = remapkeyPower;
    DriverObject->MajorFunction [IRP_MJ_PNP]  = remapkeyPnP;
    DriverObject->DriverUnload = DriverUnload;
    DriverObject->DriverExtension->AddDevice = remapkeyAddDevice;

    return STATUS_SUCCESS;
}